﻿using System;
using System.Text;
using System.Windows.Forms;
using FarPoint.Win.Spread;
using GrapeCity.CalcEngine;
using GrapeCity.Spreadsheet;

namespace SpreadNETZetaCalculator
{
    /// <summary>
    /// Spread.NET 12 Windows Forms ZETA Function Calcultor / Non-trivial Root Finder 
    /// </summary>
    public partial class ZetaCalculatorForm : Form
    {
        /// <summary>
        /// Create the ZetaCalculatorForm
        /// </summary>
        public ZetaCalculatorForm()
        {
            InitializeComponent();
            fpSpread1.AddCustomFunction(new ZetaFunction());
            formulaTextBox1.Attach(fpSpread1);
            fpSpread1.StatusBarVisible = true;
        }
        /// <summary>
        /// Returns the input values for the Zeta calculation.
        /// </summary>
        /// <param name="initval">The starting value to use for the imaginary component (the real component used is always 0.5)</param>
        /// <param name="incr">The increment to use for calculating successive values</param>
        /// <param name="rc">The number of values of Zeta to calculate</param>
        /// <param name="iter">The number of iterations to execute the calculation -- higher values will result in greater precision but take more time</param>
        /// <returns>True if the input values are valid and successfully retrieved, False otherwise</returns>
        private bool GetInputValues(ref double initval, ref double incr, ref int rc, ref long iter)
        {
            if (!double.TryParse(textBoxStart.Text, out initval))
            {
                MessageBox.Show("Please enter a double value for Start Value.");
                textBoxStart.Focus();
                return false;
            }
            if (!double.TryParse(textBoxIncrement.Text, out incr))
            {
                MessageBox.Show("Please enter a double value for Increment.");
                textBoxIncrement.Focus();
                return false;
            }
            if (!int.TryParse(textBoxCount.Text, out rc))
            {
                MessageBox.Show("Please enter an integer value for Count.");
                textBoxCount.Focus();
                return false;
            }
            long.TryParse(textBoxIter.Text, out iter);
            return true;
        }
        /// <summary>
        /// Returns the number format speifying the number of decimal places used for increment.
        /// </summary>
        /// <returns>
        /// The number format specifying the number of decimal places used for increment.
        /// </returns>
        /// <remarks>
        /// The calculator formats the first column using the number of decimal places for increment.
        /// </remarks>
        private string GetIncrementNumberFormat()
        {
            double incr = 0.1;
            if (!double.TryParse(textBoxIncrement.Text, out incr))
                return "0.0";
            StringBuilder sb = new StringBuilder();
            sb.Append("0.");
            while (incr < 1)
            {
                sb.Append("0");
                incr *= 10;
            }
            return sb.ToString();
        }
        /// <summary>
        /// Handle the "Calculate" button click.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <remarks>
        /// This event handler reinitialized the worksheet with the newly-generated Zeta calculation table.
        /// This code sets the formula to calculate ZETA as an array formula, and the spreadsheet calculates the result
        /// once, and sets the result to the range all at once using an array.
        /// The actual calculation is handled inside the custom function that is active (ZetaFunction or ZetaFunctionInfo).
        /// The title bar text is updated at the end with timing and call count information about the custom function execution.
        /// </remarks>
        private void buttonCalculate_Click(object sender, EventArgs e)
        {
            double initval = 0.0;
            double incr = 0.01;
            int rc = 12000;
            long iter = 10000;
            if (!GetInputValues(ref initval, ref incr, ref rc, ref iter))
                return;
            fpSpread1.SuspendLayout();
            Cursor = Cursors.WaitCursor;
            Text = "Calculating, please wait...";
            // inialize sheet with Zeta calculation table headers
            SheetView sheet = fpSpread1.ActiveSheet;
            sheet.Reset();
            sheet.RowCount = rc + 1;
            sheet.ColumnCount = 5;
            sheet.ReferenceStyle = FarPoint.Win.Spread.Model.ReferenceStyle.R1C1;
            sheet.Cells[0, 0].Text = "bi";
            sheet.Cells[0, 1].Text = "Re( ζ(0.5+bi) )";
            sheet.Cells[0, 2].Text = "Im( ζ(0.5+bi) )";
            sheet.Cells[0, 3].Text = "ζ(0.5+bi)";
            sheet.Cells[0, 4].Text = "IMABS( ζ(0.5+bi) )";
            sheet.Cells[0, 0, 0, 4].HorizontalAlignment = CellHorizontalAlignment.Center;
            // apply built-in styles for table header cells
            IWorkbook workbook = fpSpread1.AsWorkbook();
            workbook.ActiveSheet.Cells[0, 0, 0, 4].ApplyStyle(BuiltInStyle.Percent20Accent3);
            workbook.ActiveSheet.Cells[0, 0, 0, 4].ApplyStyle(BuiltInStyle.Heading3);
            // initialize the values to be calculated
            for (int r = 1; r <= rc; r++)
                sheet.Cells[r, 0].Value = initval + incr * r;
            // format the cells to show the number decimal places appropriate for increment
            fpSpread1.AsWorkbook().Worksheets[0].Cells[1, 0, rc, 0].NumberFormat = GetIncrementNumberFormat();
            // create the formula to calculate the ZETA function along the critical line Re = 0.5 for the 
            // set of points starting with the initial value and using the specified increment and iterations
            string formula = "ZETA(0.5,RC1," + rc.ToString().Trim() + "," + incr.ToString().Trim() + "," + iter.ToString().Trim() + ")";
            sheet.Cells[1, 1, rc, 2].FormulaArray = formula;
            // apply the COMPLEX function to the two values returned for the real (in column 1) and imaginary (in column 2) parts to get the resulting complex number calculated
            sheet.Cells[1, 3, rc, 3].Formula = "COMPLEX(RC2,RC3)";
            // apply IMABS function to resulting complex number -- this is equivalent to SQRT(RC2*RC2+RC3*RC3) and computes the length of the complex vector
            // IMABS values less than increment are considered "close" to a root and highlighted if the value is a local minimum
            sheet.Cells[1, 4, rc, 4].Formula = "IMABS(RC4)";
            // set column widths
            sheet.SetColumnWidth(0, (int)sheet.GetPreferredColumnWidth(0));
            sheet.SetColumnWidth(1, 150);
            sheet.SetColumnWidth(2, 150);
            sheet.SetColumnWidth(3, (int)sheet.GetPreferredColumnWidth(3));
            sheet.SetColumnWidth(4, (int)sheet.GetPreferredColumnWidth(4));
            // enable auto filter on the range
            workbook.ActiveSheet.Cells[0, 0, rc, 4].AutoFilter();
            // freeze first row from scrolling to keep table headers visible
            sheet.FrozenRowCount = 1;
            // renumber row headings 
            sheet.StartingRowNumber = 0;
            // clear the text in the first row header
            sheet.RowHeader.Cells[0, 0].Text = " ";
            // map the clipboard shortcut keys to special actions that copy only visible values as text
            // (so the sheet can be filtered and copy/paste works as expected)
            fpSpread1.GetInputMap(InputMapMode.WhenFocused).Put(new Keystroke(Keys.C, Keys.Control), SpreadActions.ClipboardCopyAsStringSkipHidden);
            fpSpread1.GetInputMap(InputMapMode.WhenFocused).Put(new Keystroke(Keys.Insert, Keys.Control), SpreadActions.ClipboardCopyAsStringSkipHidden);
            // create a conditional format to highlight the root (local minima IMABS < increment) values with yellow
            string conditionformula = "AND(RC5<" + incr.ToString() + ",RC5<R[1]C5,RC5<R[-1]C5)";
            FormulaConditionalFormattingRule formulaRule = new FormulaConditionalFormattingRule(conditionformula);
            formulaRule.BackColor = System.Drawing.Color.Yellow;
            sheet.SetConditionalFormatting(1, 0, rc, 5, new IConditionalFormattingRule[] { formulaRule });
            // cleanup, update window text with function call count and timing stats, enable filter roots button'
            fpSpread1.ResumeLayout();
            Cursor = Cursors.Default;
            if (checkBoxFunctionInfo.Checked)
                Text = "Zeta Calculation Complete! [ZetaFunctionInfo Call #" + ZetaFunctionInfo.cnt.ToString() + " in " + (ZetaFunctionInfo.t2 - ZetaFunctionInfo.t1).ToString() + "]";
            else
                Text = "Zeta Calculation Complete! [ZetaFunction Call #" + ZetaFunction.cnt.ToString() + " in " + (ZetaFunction.t2 - ZetaFunction.t1).ToString() + "]";
            buttonFilterRootsOnly.Enabled = true;
        }
        /// <summary>
        /// Apply a filter on the range to show only the root values.
        /// Creates a color filter for the first column in the worksheet.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonFilterRootsOnly_Click(object sender, EventArgs e)
        {
            IWorkbook workbook = fpSpread1.AsWorkbook();
            IWorksheet worksheet = workbook.ActiveSheet;
            Fill fill = new Fill(GrapeCity.Spreadsheet.Color.FromKnownColor(GrapeCity.Core.KnownColor.Yellow));
            worksheet.Cells[0, 0, worksheet.RowCount - 1, 4].AutoFilter(0, fill);
        }
        /// <summary>
        /// Toggles between using the new-style calculation function ZetaFunction and the old-style calculation function ZetaFunctionInfo.
        /// The new-style calculation function is recommended as it is slightly faster and easier to use than the old-style calculation function,
        /// which must be wrapped internally using GrapeCity.CalcEngine APIs to work in the new GrapeCity.CalcEngine.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void checkBoxFunctionInfo_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxFunctionInfo.Checked)
                // this method is marked obsolete and deprecated -- you should port your old FunctionInfo class to a Function class
                fpSpread1.ActiveSheet.AddCustomFunction(new ZetaFunctionInfo());
            else
                fpSpread1.AddCustomFunction(new ZetaFunction());
        }
        /// <summary>
        /// Event handler for the LeaveCell event.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <remarks>
        /// This event handler enables and disables the GoalSeek button based on whether the active row is a highlighted root value.
        /// </remarks>
        private void fpSpread1_LeaveCell(object sender, LeaveCellEventArgs e)
        {
            IWorkbook workbook = fpSpread1.AsWorkbook();
            // get the interior color of the cell becoming active, enable goal seek button if it's a highlighted root value
            GrapeCity.Spreadsheet.Color c = workbook.ActiveSheet.Cells[e.NewRow, e.NewColumn].DisplayFormat.Interior.Color;
            if (c.Equals(GrapeCity.Spreadsheet.Color.FromKnownColor(GrapeCity.Core.KnownColor.Yellow)))
                buttonGoalSeek.Enabled = true;
            else
                buttonGoalSeek.Enabled = false;
        }
        /// <summary>
        /// Do a high-precision goal seek to 4 decimal places
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <remarks>
        /// The spreadsheet has built-in GoalSeek logic, but that logic is limited to 2 decimal places.
        /// This function uses the CalculationEngine to perform a goal seek to 4 decimal places.
        /// </remarks>
        private void buttonGoalSeek_Click(object sender, EventArgs e)
        {
            DateTime t2, t1 = DateTime.Now; // for timing the goal-seek
            IWorkbook wb = fpSpread1.AsWorkbook();
            CalculationEngine calcengine = wb.WorkbookSet.CalculationEngine;
            IWorksheet ws = wb.ActiveSheet;
            SheetView sheet = fpSpread1.ActiveSheet;
            double initval = 0.0;
            double incr = 0.01;
            int rc = 12000;
            long iter = 10000;
            int row = sheet.ActiveRowIndex;
            int cnt, cnt2; // for counting the number of ZETA function calls performed during goal-seek
            double maxchange; // current decimal place of search
            string formula, formulaNext; // formula for value and formula for next value to be tested
            IValue result, resultNext; // calculated result of formula for value and formula for next value
            // temp variables for calculating IMABS of result and resultNext
            double temp, temp1, temp2, tempNext, tempNext1, tempNext2;
            bool nextDecimalPlace; // flag for moving to next decimal place
            double value; // current value being tested
            // save initial ZETA call count
            if (checkBoxFunctionInfo.Checked) 
                cnt = ZetaFunctionInfo.cnt;
            else
                cnt = ZetaFunction.cnt;
            if (!GetInputValues(ref initval, ref incr, ref rc, ref iter))
                return;
            // force 100K iterations for improved accuracy
            iter = 100000;
            // do goal seek on the cell [row, 0]:
            value = Math.Round((double)ws.Cells[row, 0].Value, 1) + 0.04; // start with value XXX.X4
            ws.Cells[row, 0].NumberFormat = "0.0000"; // format cell for 4 decimal places
            // my implementation for goal seek:
            Cursor = Cursors.WaitCursor;
            for (maxchange = 0.01; maxchange > 1e-5; maxchange *= 0.1)
            { // start changing the hundredths place, then move on to each sucessive decimal place until maxchange = 1e-10
                nextDecimalPlace = false; // reset flag 
                // calculate value
                formula = "ZETA(0.5," + value.ToString().Trim() + ",1,0," + iter.ToString().Trim() + ")";
                result = calcengine.Evaluate(formula, ws, new GrapeCity.Spreadsheet.CellReference(row, 1));
                temp1 = result.GetValuesArray()[0, 0].GetNumber();
                temp2 = result.GetValuesArray()[0, 1].GetNumber();
                // calculate absolute values of results
                temp = Math.Sqrt(temp1 * temp1 + temp2 * temp2);
                do
                { // next value (value + maxchange) using CalculationEngine.Evaluate
                    formulaNext = "ZETA(0.5," + (value + maxchange).ToString().Trim() + ",1,0," + iter.ToString().Trim() + ")";
                    resultNext = calcengine.Evaluate(formulaNext, ws, new GrapeCity.Spreadsheet.CellReference(row, 1));
                    tempNext1 = resultNext.GetValuesArray()[0, 0].GetNumber();
                    tempNext2 = resultNext.GetValuesArray()[0, 1].GetNumber();
                    // calculate absolute values of results
                    tempNext = Math.Sqrt(tempNext1 * tempNext1 + tempNext2 * tempNext2);
                    if (tempNext < temp)
                    {
                        value += maxchange; // next value is closer, increment value and continue checking next higher value
                        temp = tempNext;
                    }
                    else
                    { // next value is not closer, check previous value(s)
                        do
                        { // calculate previous value
                            formulaNext = "ZETA(0.5," + (value - maxchange).ToString().Trim() + ",1,0," + iter.ToString().Trim() + ")";
                            resultNext = calcengine.Evaluate(formulaNext, ws, new GrapeCity.Spreadsheet.CellReference(row, 1));
                            tempNext1 = resultNext.GetValuesArray()[0, 0].GetNumber();
                            tempNext2 = resultNext.GetValuesArray()[0, 1].GetNumber();
                            // calculate absolute value
                            tempNext = Math.Sqrt(tempNext1 * tempNext1 + tempNext2 * tempNext2);
                            if (tempNext < temp)
                            { // previous value is closer, decrement value and continue checking next previous value
                                value -= maxchange; 
                                temp = tempNext;
                            }
                            else
                            { // previous value is not closer, current value is closest -- continue to next decimal place
                                nextDecimalPlace = true; 
                                // decrement current decimal place value by 1 to avoid rounding errors, then make the new last decimal place a 4
                                value = value + 4 * (maxchange / 10);
                            }
                        } while (!nextDecimalPlace);
                    }
                } while (!nextDecimalPlace);
            }
            // discard last digit
            value = Math.Truncate(value * 10000) / 10000;
            // cleanup
            t2 = DateTime.Now;
            Cursor = Cursors.Default;
            if (checkBoxFunctionInfo.Checked)
                cnt2 = ZetaFunctionInfo.cnt;
            else
                cnt2 = ZetaFunction.cnt;
            sheet.Cells[row, 0].Value = value;
            sheet.Columns[0].Width = sheet.GetPreferredColumnWidth(0);
            MessageBox.Show("Zero Value: " + value.ToString() + "\r\nCalculation completed in " + (t2 - t1).ToString() + " with " + (cnt2 - cnt).ToString() + " calls to ZETA function.");
        }
    }
}
